//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "paca.h"
#include <math.h>
#include <algorithm>
#include "ASMVars.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
// Pantoghraphic Cellular Automata Inpainting
//
// Core class
//
// Author: Sinia Petri
// You may freely distribute, copy, cut and paste this code, as long as...whatever.
//---------------------------------------------------------------------------
__fastcall TPaca::TPaca(TIEBitmap *src, TIEBitmap *dst)
{
// create all necessary images and surfaces
randomize();
std::srand(time(NULL));
//
executionPass = 0;
imgWidth  = src->Width;
imgHeight = src->Height;
imgWidthM1  = imgWidth - 1;
imgHeightM1 = imgHeight - 1;
// create surfaces and control images
srcSurface = new TSpadeSurfaceIen(src);
dstSurface = new TSpadeSurfaceIen(dst);
// control input map
controlInput = new TIEBitmap(imgWidth, imgHeight, ie8g);
controlInput->Fill(0);
controlInputSurface = new TSpadeSurfaceIen(controlInput);
// copy control inp to out
controlOutput = new TIEBitmap(controlInput);
controlOutputSurface = new TSpadeSurfaceIen(controlOutput);
//
controlInputSurface->Options.PixelOverFlow = eopSkip;
controlOutputSurface->Options.PixelOverFlow = eopSkip;
//
// paca map
createPacaMap();
//
maskSurface = 0;
targetSite = 0;
sourceSite = 0;
targetDistrib = 0;
}
//---------------------------------------------------------------------------
// destructor
__fastcall TPaca::~TPaca(void)
{
deleteSites();
delete srcSurface;
delete dstSurface;
delete controlInputSurface;
delete controlInput;
delete controlOutputSurface;
delete controlOutput;
deletePacaMap();
if (maskSurface)
   delete maskSurface;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::createPacaMap(void)
{
// paca map
pacaMap.X = new int *[imgHeight];
pacaMap.Y = new int *[imgHeight];
pacaMap.Iteration = new int *[imgHeight];
pacaMap.SiteIndex = new int *[imgHeight];
pacaMap.InpStrength = new float *[imgHeight];
pacaMap.OutStrength = new float *[imgHeight];
for (int i = 0; i < imgHeight; i++)
	{
	pacaMap.X[i] = new int [imgWidth];
	pacaMap.Y[i] = new int [imgWidth];
	pacaMap.Iteration[i] = new int [imgWidth];
	pacaMap.SiteIndex[i] = new int [imgWidth];
	pacaMap.InpStrength[i] = new float [imgWidth];
	pacaMap.OutStrength[i] = new float [imgWidth];
	for (int j = 0; j < imgWidth; j++)
		{
		pacaMap.X[i][j] = 0;;
		pacaMap.Y[i][j] = 0;
		pacaMap.Iteration[i][j] = 0;
		pacaMap.SiteIndex[i][j] = -1;
		pacaMap.InpStrength[i][j] = 0.0;
		pacaMap.OutStrength[i][j] = 0.0;
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TPaca::deletePacaMap(void)
{
for (int i = 0; i < imgHeight; i++)
	{
	delete []pacaMap.X[i];
	delete []pacaMap.Y[i];
	delete []pacaMap.Iteration[i];
	delete []pacaMap.SiteIndex[i];
	delete []pacaMap.InpStrength[i];
	delete []pacaMap.OutStrength[i];
	}
delete []pacaMap.X;
delete []pacaMap.Y;
delete []pacaMap.Iteration;
delete []pacaMap.SiteIndex;
delete []pacaMap.InpStrength;
delete []pacaMap.OutStrength;
}
//---------------------------------------------------------------------------
// prepare
//---------------------------------------------------------------------------
void __fastcall TPaca::Prepare(TIEBitmap *mask, TIEBitmap *hardMask, TRect *pSelRect)
{
int i, j;
selRect = pSelRect;
if (maskSurface)
   delete maskSurface;
maskSurface = new TSpadeSurfaceIen(mask);
targetRect.left  = selRect->left;
targetRect.right = selRect->right;
targetRect.top   = selRect->top;
targetRect.bottom = selRect->bottom;
// If selection area is too small...shit happens. There is no pixel with 0xff.
// Do not use small selections with big feather. Instead, use selection boundary smoothing
hardMaskByte = 0xff;
if (hardMask)
   {
   TSpadeSurfaceBasic *hardMaskSurface = new TSpadeSurfaceIen(hardMask);
   for (i = targetRect.top; i < targetRect.bottom; i++)
	   {
	   for (j = targetRect.left; j < targetRect.right; j++)
		   {
		   if (hardMaskSurface->DirectByte[i][j] == 0xff && maskSurface->DirectByte[i][j] < hardMaskByte)
			  hardMaskByte = maskSurface->DirectByte[i][j];
		   }
	   }
   delete hardMaskSurface;
   }
// reset control surface
resetControlSurface(CTRL_UNMASKED | CTRL_MASKED | CTRL_PARTIALLY_MASKED);
//resetControlSurface(CTRL_UNMASKED | CTRL_MASKED);
fMaxStrength = 1.0;
maxProbability = 1;
// reset pacaMap
for (i = 0; i < imgHeight; i++)
	{
	for (j = 0; j < imgWidth; j++)
		{
		pacaMap.X[i][j] = 0;;
		pacaMap.Y[i][j] = 0;
		pacaMap.Iteration[i][j] = 0;
		pacaMap.SiteIndex[i][j] = -1;
		pacaMap.InpStrength[i][j] = 0.0;
		pacaMap.OutStrength[i][j] = 0.0;
		}
	}
}
//---------------------------------------------------------------------------
// Seeding methods
//---------------------------------------------------------------------------
bool __fastcall TPaca::Seed(int pSeedStyle, int pDivision, int pPushValue, int pPullValue, bool pAutoAdjustPull)
{
// seed as required
pushValue = pPushValue;
pullValue = pPullValue;
seedObj = new TSeed(maskSurface, controlInputSurface, &targetRect, hardMaskByte);
if (!seedObj->Seed(pSeedStyle, pDivision))
   {
   delete seedObj;
   seedObj = 0;
   return false;
   }
siteCount = seedObj->SeedList->Count;
averageAdjust = pullValue;
seedTarget(pullValue, pullValue, pAutoAdjustPull);
seedSource(pushValue, pushValue);
validateSites(true);
delete seedObj;
seedObj = 0;
return true;
//return false;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::ReseedManual(TPoint *targetSeed, TPoint *sourceSeed)
{
// reseed sites according to manual adjustment
int i, j, k;
for (i = 0; i < imgHeight; i++)
	{
	for (j = 0; j < imgWidth; j++)
		pacaMap.SiteIndex[i][j] = -1;
	}
for (k = 0; k < siteCount; k++)
	{
	targetSite[k].x = targetSeed[k].x;
	targetSite[k].y = targetSeed[k].y;
	sourceSite[k].x = sourceSeed[k].x;
	sourceSite[k].y = sourceSeed[k].y;
	}
validateSites(true);
return true;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::OnionPeel(void)
{
// Usually, when mirroring is not checked, and if pull or push value is small
// we need onion peeling to complete inpainting
//copyDst2Src(); // test
resetControlSurface(CTRL_UNMASKED);
//resetControlSurface(CTRL_UNMASKED | CTRL_PARTIALLY_MASKED);
int wX, wY;
int validSites = 0;
for (int k = 0; k < siteCount; k++)
	{
	wY = targetSite[k].y;
	wX = targetSite[k].x;
	//if (walkAround(&wY, &wX, targetSite[k].inDirection, CTRL_TARGET | CTRL_UNMASKED | CTRL_SOURCE, 0, 0))
	if (walkAround(&wY, &wX, targetSite[k].inDirection, CTRL_TARGET, pullValue, pullValue))
	   {
	   targetSite[k].y = wY;
	   targetSite[k].x = wX;
	   controlInputSurface->DirectByte[wY][wX] = CTRL_TARGET;
	   validSites++;

	   }
	else
	   {
	   targetSite[k].valid = false;
	   sourceSite[k].valid = false;
	   }
	}
if (validSites == 0)
   return true;
validateSites(false);
return false;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::deleteSites(void)
{
// clear all sites
if (targetSite)
   {
   delete []targetSite;
   targetSite = 0;
   }
if (sourceSite)
   {
   delete []sourceSite;
   sourceSite = 0;
   }
if (targetDistrib)
   {
   delete []targetDistrib;
   targetDistrib = 0;
   }
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::walkAround(int *y, int *x, TWalkDirection direction, BYTE cByte, int addY, int addX)
{
// walk around in given direction
int y_incr = 0, x_incr = 0;
switch (direction)
	{
	case wUp:
		 y_incr = -1;
		 break;
	case wDown:
		 y_incr = +1;
		 break;
	case wLeft:
		 x_incr = -1;
		 break;
	case wRight:
		 x_incr = +1;
		 break;
	case wUpLeft:
		 y_incr = -1;
		 x_incr = -1;
		 break;
	case wUpRight:
		 y_incr = -1;
		 x_incr = +1;
		 break;
	case wDownLeft:
		 y_incr = +1;
		 x_incr = -1;
		 break;
	case wDownRight:
		 y_incr = +1;
		 x_incr = +1;
		 break;
	}
//while (maskSurface->DirectByte[*y][*x] == 0xff)
while (controlInputSurface->DirectByte[*y][*x] & cByte) // walk forward
	{
	*x += x_incr;
	*y += y_incr;
	if (!controlInputSurface->SetSafeIndex(y, x)) // back one step
	   {
	   *x -= x_incr;
	   *y -= y_incr;
	   return false;
	   }
	}
// add value
*x += x_incr * addX;
*y += y_incr * addY;
while (!controlInputSurface->SetSafeIndex(y, x)) // walk back if beyond image size
   {
   *x -= x_incr;
   *y -= y_incr;
   return false;
   }
return true;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::translateAround(int *y, int *x, TWalkDirection direction, BYTE cByte, int addY, int addX)
{
// translate site in given direction
int y_incr = 0, x_incr = 0;
switch (direction)
	{
	case wUp:
		 y_incr = -1;
		 break;
	case wDown:
		 y_incr = +1;
		 break;
	case wLeft:
		 x_incr = -1;
		 break;
	case wRight:
		 x_incr = +1;
		 break;
	case wUpLeft:
		 y_incr = -1;
		 x_incr = -1;
		 break;
	case wUpRight:
		 y_incr = -1;
		 x_incr = +1;
		 break;
	case wDownLeft:
		 y_incr = +1;
		 x_incr = -1;
		 break;
	case wDownRight:
		 y_incr = +1;
		 x_incr = +1;
		 break;
	}
*x += x_incr * addX;
*y += y_incr * addY;
while (!controlInputSurface->SetSafeIndex(y, x)) // walk back if beyond image size
   {
   *x -= x_incr;
   *y -= y_incr;
   }
while (controlInputSurface->DirectByte[*y][*x] & cByte) // walk backward
	{
   *x -= x_incr;
   *y -= y_incr;
   // set safe index???
   }
}
//---------------------------------------------------------------------------
void __fastcall TPaca::resetControlSurface(BYTE reset)
{
// reset input and output control surface (grayscale images)
if (reset & CTRL_MASKED)
   targetArea = 0;
for (int i = 0; i < imgHeight; i++)
	{
	for (int j = 0; j < imgWidth; j++)
		{
		if (maskSurface->DirectByte[i][j] == 0x00 && (reset & CTRL_UNMASKED))
		   controlInputSurface->DirectByte[i][j] = CTRL_UNMASKED;
		else
		   {
		   targetArea++;
		   if (maskSurface->DirectByte[i][j] >= hardMaskByte && (reset & CTRL_MASKED))	// fully masked
			  controlInputSurface->DirectByte[i][j] = CTRL_MASKED;
		   else if (reset & CTRL_PARTIALLY_MASKED) // partially masked
		   	  controlInputSurface->DirectByte[i][j] = CTRL_PARTIALLY_MASKED;
		   }
		controlOutputSurface->DirectByte[i][j] = controlInputSurface->DirectByte[i][j];
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TPaca::seedTarget(int addY, int addX, bool autoAdjustPull)
{
// seed target sites
targetSite = new structSite[siteCount];
targetDistrib = new int[siteCount];
int y, x, averCount = 0;
averageAdjust = 0;
for (int k = 0; k < siteCount; k++)
	{
	//targetSite[siteCount].min_quadrance = 99999999;
	TSeedPoint *outPt = (TSeedPoint*)(seedObj->SeedList->Items[k]);
	targetSite[k].x = outPt->x;
	targetSite[k].y = outPt->y;
	targetSite[k].inDirection = outPt->sweepDirection;
	targetSite[k].outDirection = outPt->walkDirection;
	y = targetSite[k].y;
	x = targetSite[k].x;
	if (autoAdjustPull)
	   {
	   int wX, wY, xLen, yLen;
	   wY = y;
	   wX = x;
	   walkAround(&wY, &wX, targetSite[k].inDirection, CTRL_TARGET | CTRL_MASKED, 0, 0);
	   xLen = abs(x - wX);
	   yLen = abs(y - wY);
	   addX = (xLen>>1); //(pushValue>>1);
	   addY = (yLen>>1); //(pushValue>>1);
	   if (addX < 0)
		  addX = (xLen>>1);
	   if (addY < 0)
		  addY = (yLen>>1);
	   }
	// show adjustment - just for information
	if (addX > 0)
	   {
	   averageAdjust += addX;
	   averCount++;
	   }
	if (addY > 0)
	   {
	   averageAdjust += addY;
	   averCount++;
	   }
	translateAround(&y, &x, targetSite[k].inDirection, CTRL_TARGET | CTRL_PARTIALLY_MASKED | CTRL_UNMASKED | CTRL_SOURCE, addY, addX);
	//translateAround(&y, &x, targetSite[k].inDirection, CTRL_TARGET | CTRL_UNMASKED | CTRL_SOURCE, addY, addX);
	if (controlInputSurface->SetSafeIndex(&y, &x))
	   {
	   targetSite[k].y = y;
	   targetSite[k].x = x;
	   }
	controlInputSurface->DirectByte[targetSite[k].y][targetSite[k].x] = CTRL_TARGET;
	}
if (averCount)
   averageAdjust = fast_ftol((float)averageAdjust / averCount);
}
//---------------------------------------------------------------------------
void __fastcall TPaca::seedSource(int addY, int addX)
{
// seed source sites
sourceSite = new structSite[siteCount];
// loop through target sites
int y, x;
for (int k = 0; k < siteCount; k++)
	{
	y = targetSite[k].y;
	x = targetSite[k].x;
	walkAround(&y, &x, targetSite[k].outDirection, CTRL_MASKED | CTRL_TARGET | CTRL_PARTIALLY_MASKED, addY, addX);
	//walkAround(&y, &x, targetSite[k].outDirection, CTRL_MASKED | CTRL_TARGET, addY, addX);
	sourceSite[k].y = y;
	sourceSite[k].x = x;
	targetSite[k].pair_y = y;
	targetSite[k].pair_x = x;
	}
}
//---------------------------------------------------------------------------
void __fastcall TPaca::validateSites(bool removeTarget)
{
// validate sites
int retries;
int i, x, y;
bool validSite = true;
if (removeTarget)
   resetControlSurface(CTRL_MASKED);	// remove CTRL_TARGET marks
sourceRect.left = imgWidth;
sourceRect.right = 0;
sourceRect.top = imgHeight;
sourceRect.bottom = 0;
for (i = 0; i < siteCount; i++)
	{
	x = targetSite[i].x;
	y = targetSite[i].y;
	if (controlInputSurface->SetSafeIndex(&y, &x))
	   {
	   if (controlInputSurface->DirectByte[y][x] == (CTRL_MASKED | CTRL_TARGET))
		  validSite = true;
	   }
	else
	   validSite = false;
	x = sourceSite[i].x;
	y = sourceSite[i].y;
	if (controlInputSurface->SetSafeIndex(&y, &x))
	   {
		 if (controlInputSurface->DirectByte[y][x] == CTRL_UNMASKED)
			validSite &= true;
	   }
	else
	   validSite &= false;
	targetSite[i].valid = validSite;
	sourceSite[i].valid = validSite;
	int tx = targetSite[i].x;
	int ty = targetSite[i].y;
	if (validSite)
	   {
	   sourceSite[i].x = x;
	   sourceSite[i].y = y;
	   targetSite[i].pair_x = x;
	   targetSite[i].pair_y = y;
	   controlInputSurface->DirectByte[y][x] = CTRL_SOURCE;
	   pacaMap.X[ty][tx] = x;	// source x
	   pacaMap.Y[ty][tx] = y;	// source y
	   pacaMap.SiteIndex[ty][tx]   = i;
	   pacaMap.InpStrength[ty][tx] = OPS_INIT_STRENGTH;  // initial target strength
	   //pacaMap.InpStrength[y][x]   = OPS_INIT_STRENGTH;  // initial source strength
	   pacaMap.OutStrength[ty][tx] = OPS_INIT_STRENGTH;  // initial target strength
	   //pacaMap.OutStrength[y][x]   = OPS_INIT_STRENGTH;  // initial source strength
	   controlInputSurface->DirectByte[ty][tx] = CTRL_TARGET;
	   // get sourceRect pts
	   if (y < sourceRect.top)
		  sourceRect.top = y;
	   if (y > sourceRect.bottom)
		  sourceRect.bottom = y;
	   if (x < sourceRect.left)
		  sourceRect.left = x;
	   if (x > sourceRect.right)
		  sourceRect.right = x;
	   }
	else
	   controlInputSurface->DirectByte[ty][tx] = CTRL_MASKED;
	}
// adjust source rect
sourceRect.left   -= targetRect.Width();
sourceRect.right  += targetRect.Width();
sourceRect.top    -= targetRect.Height();
sourceRect.bottom += targetRect.Height();
if (sourceRect.left < 0)
   sourceRect.left = 0;
if (sourceRect.right > imgWidth)
   sourceRect.right = imgWidth;
if (sourceRect.top < 0)
   sourceRect.top = 0;
if (sourceRect.bottom > imgHeight)
   sourceRect.bottom = imgHeight;
}
//---------------------------------------------------------------------------
// Execution methods
//---------------------------------------------------------------------------
bool __fastcall TPaca::Execute(int maxIterations, bool mirror, int pOutputPixelMode, int pDistanceMode,
							  int *pCreationOps, int pMutateIndex, int pMutateStep, bool pVarProbability)
{
// set private variables and choos methods according to imput parameters.
bool lRet = false;
varProbability = pVarProbability;
if (mirror)
   {
   dispN33I = &N33I[0];
   //dispN33I = &mirrN33I[0];
   dispN33J = &mirrN33J[0];
   //dispN33J = &N33J[0];
   }
else
   {
   dispN33I = &N33I[0];
   dispN33J = &N33J[0];
   }
switch (pOutputPixelMode)
  {
  case 0:   // convolve
	   doPixelOutput = setOutputPixelConvolve;
	   //doPixelOutput = setOutputPixelAverage;
	   break;
  case 1:	// aweighted sum
	   doPixelOutput = setOutputPixelWeightedSum;
	   break;
  case 2:   // stronger wins
	   doPixelOutput = setOutputPixelStrongest;
	   break;
  default:
	   doPixelOutput = setOutputPixelConvolve;
	   break;
  }
switch(pDistanceMode)
  {
  case 0:	// good ol' Euclid
	   getDistance = getEuclidDistance;
	   break;
  default:	// luminance
   	   getDistance = getLumaDistance;
	   break;
  }
mutateCreationStep = pMutateStep;
switch (pMutateIndex)
  {
  case 1:
	   mutateCreation = rotateOps;
	   break;
  case 2:
	   mutateCreation = shuffleOps;
	   break;
  default:
	   mutateCreation = nopOps;
	   break;
}
int siteCreated = true;
int i, j;
// test points - output target vs source
/*
if (MessageDlg("Show sites?", mtConfirmation, TMsgDlgButtons() << mbYes<<mbNo, 0) == mrYes)
   {
	for (i = 0; i < siteCount; i++)
		{
		if (!targetSite[i].valid)
		   continue;
		dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[0] = 0x00;
		dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[1] = 0x00;
		dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[2] = 0xff;
		}
	for (i = 0; i < siteCount; i++)
		{
		if (!sourceSite[i].valid)
		   continue;
		dstSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[0] = 0x00;
		dstSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[1] = 0xff;
		dstSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[2] = 0x00;
		}
	return true;
	}
*/
// copy initial target pixels to dst
for (i = 0; i < siteCount; i++)
	{
	if (!targetSite[i].valid)
	   continue;
	targetSite[i].BGR[0] =  srcSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[0];
	targetSite[i].BGR[1] =  srcSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[1];
	targetSite[i].BGR[2] =  srcSurface->DirectBGR[sourceSite[i].y][sourceSite[i].x].bgr[2];
	dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[0] = targetSite[i].BGR[0];
	dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[1] = targetSite[i].BGR[1];
	dstSurface->DirectBGR[targetSite[i].y][targetSite[i].x].bgr[2] = targetSite[i].BGR[2];
	}
for (i = 0; i < 9; i++)
	{
	if (i == 4)
	   continue;
	//inputCreationOps[OPS_pos[i]] = pCreationOps[i];
	//if (executionPass == 0)
	//   creationOps[OPS_pos[i]] = 0;	// always star with ORs
	creationOps[OPS_pos[i]] = pCreationOps[i];
	}
int iterations = 0;
int caIterator = 1;
targetSitesCount = 0;
while (siteCreated)
	{
	processCA();
	swapControlImages();
	// test
	if (++caIterator > mutateCreationStep)
	   {
	   mutateCreation(creationOps);
	   caIterator = 1;
	   }
	//progressCall(iterator + 1, _iterations);
	/*
	if (iterations == 0 && executionPass == 0)	// set real ops
	   {
	   for (i = 0; i < 9; i++)
		   creationOps[i] = inputCreationOps[i];
	   }
	*/
	iterations++;
	siteCreated = targetSiteCreated;
	// test
	if (iterations > maxIterations)
	   break;
	if (targetSitesCount >= targetArea)
	   {
	   lRet = true;
	   break;
	   }
	}
//progressCall(0, 0);
executionPass++;
// test - show strength map
/*
float strength, max_strength = 0;
for (i = targetRect.top; i < targetRect.bottom; i++)
	{
	for (j = targetRect.left; j < targetRect.right; j++)
		{
		if (pacaMap.OutStrength[i][j] > max_strength)
		   max_strength = pacaMap.OutStrength[i][j];
		}
	}
for (i = targetRect.top; i < targetRect.bottom; i++)
	{
	for (j = targetRect.left; j < targetRect.right; j++)
		{
		strength = pacaMap.OutStrength[i][j] * 255 / max_strength;
		dstSurface->DirectBGR[i][j].bgr[0] = (BYTE)strength;
		dstSurface->DirectBGR[i][j].bgr[1] = (BYTE)strength;
		dstSurface->DirectBGR[i][j].bgr[2] = (BYTE)strength;
		}
	}
*/
return lRet;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::nopOps(int *ops)
{
// Las luces aprendidas, pero nadie en casa
}
//---------------------------------------------------------------------------
void __fastcall TPaca::rotateOps(int *ops)
{
// rotate creation matrix
int last = ops[7];
for (int i = 7; i > 0; i--)
	ops[i] = ops[i-1];
ops[0] = last;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::shuffleOps(int *ops)
{
// shuffle creation matrix
std::random_shuffle( &ops[0], &ops[8]);
}
//---------------------------------------------------------------------------
void __fastcall TPaca::swapControlImages(void)
{
// we need to swap control surfaces with every CA iteration
controlSwap = controlInput;
controlInput = controlOutput;
controlOutput = controlSwap;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::resetLogicContainer(void)
{
for (int i = 0; i < 5; i++)
	{
	createLogic[i].Count = 0;
	createLogic[i].Container[0] = 0x01;
	createLogic[i].Result = false;
	}
}
//---------------------------------------------------------------------------
void __fastcall TPaca::addLogicContainer(structCaLogic *logic, int index, bool v1)
{
logic[index].Container[logic[index].Count++] = v1 ? 0x01 : 0x00 ;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::calculateLogicContainer(structCaLogic *logic)
{
int i;
// OR
for (i = 1; i < logic[0].Count; i++)
	logic[0].Container[0] = logic[0].Container[0] | logic[0].Container[i];
logic[0].Result = (bool)logic[0].Container[0];
// AND
for (i = 1; i < logic[1].Count; i++)
	logic[1].Container[0] = logic[1].Container[0] & logic[1].Container[i];
logic[1].Result = (bool)logic[1].Container[0];
// XOR
for (i = 1; i < logic[2].Count; i++)
	logic[2].Container[0] = logic[2].Container[0] + logic[2].Container[i];
logic[2].Result = (bool)(logic[2].Container[0] & 0x01);
// NAND
for (i = 1; i < logic[3].Count; i++)
	logic[3].Container[0] = logic[3].Container[0] + logic[3].Container[i];
logic[3].Container[0] = (logic[3].Container[0] >> 1);
logic[3].Result = !(bool)logic[3].Container[0];
// TRUE - always true
logic[4].Result = true;
// Result
return logic[0].Result && logic[1].Result && logic[2].Result && logic[3].Result && logic[4].Result;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::exploreTargetNeighbour(int i, int j)
{
int si, sj, k, neighbour_count = 0;
bool neighbour;
fCumulStrength = 0.0;
for (k = 0; k < 8; k++)
	{
	si = i + N33I[k];
	sj = j + N33J[k];
	if (!controlInputSurface->SetSafeIndex(&si, &sj))
	   continue;
	neighbour = (controlInputSurface->DirectByte[si][sj] == CTRL_TARGET);
	if (neighbour)
	   {
	   targetNbh[neighbour_count].SiteIndex = pacaMap.SiteIndex[si][sj];
	   targetNbh[neighbour_count].X = sj;
	   targetNbh[neighbour_count].Y = si;
	   targetNbh[neighbour_count].k = k;
	   int tX = pacaMap.X[si][sj];
	   int tY = pacaMap.Y[si][sj];
	   sourceNbh[neighbour_count].X = tX;
	   sourceNbh[neighbour_count].Y = tY;
	   sourceNbh[neighbour_count].k = k;
	   // source neighbours pixels -> put it in target
	   targetNbh[neighbour_count].fBGR[0] = (float)srcSurface->DirectBGR[tY][tX].bgr[0];
	   targetNbh[neighbour_count].fBGR[1] = (float)srcSurface->DirectBGR[tY][tX].bgr[1];
	   targetNbh[neighbour_count].fBGR[2] = (float)srcSurface->DirectBGR[tY][tX].bgr[2];
	   //strength value
	   targetNbh[neighbour_count].fStrength = pacaMap.InpStrength[si][sj];
	   fCumulStrength += targetNbh[neighbour_count].fStrength;
	   neighbour_count++;
	   }
	addLogicContainer(createLogic, creationOps[k], neighbour);
	}
return neighbour_count;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::exploreHoleNeighbour(int i, int j, BYTE cByte)
{
int si, sj, k, neighbour_count = 0;
bool neighbour;
targetCentral.fBGR[0] = targetCentral.fBGR[1] = targetCentral.fBGR[2] = 0.0;
targetCentral.fStrength = 0.0;
targetCentral.SiteIndex = pacaMap.SiteIndex[i][j];
for (k = 0; k < 8; k++)
	{
	si = i + N33I[k];
	sj = j + N33J[k];
	if (!maskSurface->SetSafeIndex(&si, &sj))
	   continue;
	neighbour = (controlInputSurface->DirectByte[si][sj] == cByte);
	if (neighbour)
	   {
	   targetNbh[neighbour_count].SiteIndex = pacaMap.SiteIndex[si][sj];
	   targetNbh[neighbour_count].X = sj;
	   targetNbh[neighbour_count].Y = si;
	   targetNbh[neighbour_count].k = k;
	   // neighbours pixels
	   targetNbh[neighbour_count].fBGR[0] = (float)srcSurface->DirectBGR[si][sj].bgr[0];
	   targetNbh[neighbour_count].fBGR[1] = (float)srcSurface->DirectBGR[si][sj].bgr[1];
	   targetNbh[neighbour_count].fBGR[2] = (float)srcSurface->DirectBGR[si][sj].bgr[2];
	   //strength value
	   targetNbh[neighbour_count].fStrength = pacaMap.InpStrength[si][sj];
	   if (targetNbh[neighbour_count].fStrength < OPS_TOLERANCE)
		  {
		  // ups!
		  int mate = 23;
		  }
	   neighbour_count++;
	   }
	}
return neighbour_count;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::exploreSiteNeighbour(int i, int j, BYTE cByte)
{
int si, sj, k, neighbour_count = 0;
bool neighbour;
int tX = pacaMap.X[i][j];
int tY = pacaMap.Y[i][j];
targetCentral.fBGR[0] = (float)srcSurface->DirectBGR[tY][tX].bgr[0];
targetCentral.fBGR[1] = (float)srcSurface->DirectBGR[tY][tX].bgr[1];
targetCentral.fBGR[2] = (float)srcSurface->DirectBGR[tY][tX].bgr[2];
targetCentral.fStrength = pacaMap.InpStrength[i][j];
outSiteIndex = pacaMap.SiteIndex[i][j];
for (k = 0; k < 8; k++)
	{
	si = i + N33I[k];
	sj = j + N33J[k];
	if (!controlInputSurface->SetSafeIndex(&si, &sj))
	   continue;
	neighbour = (controlInputSurface->DirectByte[si][sj] == cByte);
	if (neighbour)
	   {
	   targetNbh[neighbour_count].SiteIndex = pacaMap.SiteIndex[si][sj];
	   targetNbh[neighbour_count].X = sj;
	   targetNbh[neighbour_count].Y = si;
	   targetNbh[neighbour_count].k = k;
	   // neighbours pixels
	   targetNbh[neighbour_count].fBGR[0] = (float)srcSurface->DirectBGR[si][sj].bgr[0];
	   targetNbh[neighbour_count].fBGR[1] = (float)srcSurface->DirectBGR[si][sj].bgr[1];
	   targetNbh[neighbour_count].fBGR[2] = (float)srcSurface->DirectBGR[si][sj].bgr[2];
	   //strength value
	   targetNbh[neighbour_count].fStrength = pacaMap.InpStrength[si][sj];
	   //if (fnbhStrength[neighbour_count] == 0)
	   //	  {
	   //	  // ups!
	   //	  }
	   neighbour_count++;
	   }
	}
return neighbour_count;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::exploreBorderNeighbour(int i, int j, BYTE cByte)
{
int si, sj, k, neighbour_count = 0;
bool neighbour;
targetCentral.fBGR[0] = (float)srcSurface->DirectBGR[i][j].bgr[0];
targetCentral.fBGR[1] = (float)srcSurface->DirectBGR[i][j].bgr[1];
targetCentral.fBGR[2] = (float)srcSurface->DirectBGR[i][j].bgr[2];
targetCentral.fStrength = 0.0;
for (k = 0; k < 8; k++)
	{
	si = i + N33I[k];
	sj = j + N33J[k];
	if (!maskSurface->SetSafeIndex(&si, &sj))
	   continue;
	neighbour = (controlInputSurface->DirectByte[si][sj] == cByte);
	if (neighbour)
	   {
		targetNbh[neighbour_count].fBGR[0] = (float)srcSurface->DirectBGR[si][sj].bgr[0];
		targetNbh[neighbour_count].fBGR[1] = (float)srcSurface->DirectBGR[si][sj].bgr[1];
		targetNbh[neighbour_count].fBGR[2] = (float)srcSurface->DirectBGR[si][sj].bgr[2];
		neighbour_count++;
	   }
	/*
	fnbhBGR[k][0] = (float)srcSurface->DirectBGR[si][sj].bgr[0];
	fnbhBGR[k][1] = (float)srcSurface->DirectBGR[si][sj].bgr[1];
	fnbhBGR[k][2] = (float)srcSurface->DirectBGR[si][sj].bgr[2];
	if (neighbour)
	   {
	   fnbhStrength[k] = 1;
	   neighbour_count++;
	   }
	else
	   fnbhStrength[k] = 1;
	*/
	}
return neighbour_count;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::doCreation(int i, int j, int neighbour_count)
{
// creation
bool result = calculateLogicContainer(createLogic);
// variabile probability, according to site strength
if (result && varProbability)
   {
   fCumulStrength /= (float)neighbour_count;
   int cumulStrength = fast_ftol(fCumulStrength);
   if (cumulStrength != maxProbability)
	  result = result && (random(maxProbability + 1) < cumulStrength);
   }
return result;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::getSourceCoords(int neighbour_count, BYTE cByte)
{
// get coordinates of source site
bool coordsOk = true;
int index = (neighbour_count == 1) ? 0 : random(neighbour_count);
iCoordX = sourceNbh[index].X - dispN33J[sourceNbh[index].k];
iCoordY = sourceNbh[index].Y - dispN33I[sourceNbh[index].k];
coordsOk = controlOutputSurface->SetSafeIndex(&iCoordY, &iCoordX);
if (coordsOk)
   {
   if (controlOutputSurface->DirectByte[iCoordY][iCoordX] & cByte)
	  return true;
   }
if (neighbour_count > 1)
   {
   // scan all
   for (int i = 0; i < neighbour_count; i++)
	   {
	   iCoordX = sourceNbh[i].X - dispN33J[sourceNbh[i].k];
	   iCoordY = sourceNbh[i].Y - dispN33I[sourceNbh[i].k];
	   coordsOk = controlOutputSurface->SetSafeIndex(&iCoordY, &iCoordX);
	   if (coordsOk)
		  {
		  if (controlOutputSurface->DirectByte[iCoordY][iCoordX] & cByte)
			 return true;
		  }
	   }
   return false;
   }
return false;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::getSiteIndex(int i, int j, int neighbour_count)
{
// There are more then one neighbour site while creating new site...which one to choose for indexing.
// Let's say the frequent one.
int site_index;
if (neighbour_count == 1)
   site_index = targetNbh[0].SiteIndex;
else
   {
   int k, max_freq;
   // naive approach
   for (k = 0; k < siteCount; k++)
	   targetDistrib[k] = 0;
   for (k = 0; k < neighbour_count; k++)
	   targetDistrib[targetNbh[k].SiteIndex]++;
   max_freq = site_index = 0;
   for (k = 0; k < siteCount; k++)
	   {
	   if (targetDistrib[k] > max_freq)
		  {
		  max_freq = targetDistrib[k];
		  site_index = k;
		  }
	   }
   }
return site_index;
}
//---------------------------------------------------------------------------
float __fastcall TPaca::getLuminance(float *fBgr)
{
// Luminance
return (fBgr[0] * OPS_LUMINANCE[0] + fBgr[1] * OPS_LUMINANCE[1] +  fBgr[2] * OPS_LUMINANCE[2] ); // / 255.0;
}
//---------------------------------------------------------------------------
float __fastcall TPaca::getLumaDistance(float *fBgr1, float *fBgr2)
{
// luminance diference
return fabs(getLuminance(fBgr1) - getLuminance(fBgr2)) + 1.0;
}
//---------------------------------------------------------------------------
float __fastcall TPaca::getEuclidDistance(float *fBgr1, float *fBgr2)
{
// Good ol' Euclid
float quadrance = (fBgr1[0] - fBgr2[0]) * (fBgr1[0] - fBgr2[0]);
quadrance += (fBgr1[1] - fBgr2[1]) * (fBgr1[1] - fBgr2[1]);
quadrance += (fBgr1[2] - fBgr2[2]) * (fBgr1[2] - fBgr2[2]);
return sqrt(quadrance + 1.0);
}
//---------------------------------------------------------------------------
void __fastcall TPaca::calculateStrength(int i, int j, int neighbour_count, float *fBGR)
{
// calculate strength of newly created site
float distance;
targetCentral.fStrength = targetNbh[0].fStrength;
if (neighbour_count == 1)
   distance    = getDistance(targetNbh[0].fBGR, fBGR);
else
  {
  float nbc = (float)neighbour_count;
  for (int k = 1; k < neighbour_count; k++)
	  {
	  targetCentral.fStrength += targetNbh[k].fStrength;
	  targetNbh[0].fBGR[0] += targetNbh[k].fBGR[0];
	  targetNbh[0].fBGR[1] += targetNbh[k].fBGR[1];
	  targetNbh[0].fBGR[2] += targetNbh[k].fBGR[2];
	  }
  targetCentral.fStrength /= nbc;
  targetNbh[0].fBGR[0] /= nbc;
  targetNbh[0].fBGR[1] /= nbc;
  targetNbh[0].fBGR[2] /= nbc;
  distance = getDistance(targetNbh[0].fBGR, fBGR);
  }
targetCentral.fStrength += 1.0/distance;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::getStrongestOne(int neighbour_count)
{
// get stongest site index
float max_strength = 0.0, max_diff_strength = 0.0;
int index_glob, index_diff;
for (int k = 0; k < neighbour_count; k++)
	{
	if (targetNbh[k].fStrength > max_strength)
	   {
	   max_strength = targetNbh[k].fStrength;
	   index_glob = k;
	   }
	if (targetNbh[k].fStrength > max_diff_strength && targetNbh[k].SiteIndex != targetCentral.SiteIndex)
	   {
	   max_diff_strength = targetNbh[k].fStrength;
	   index_diff = k;
	   }
	}
if (max_diff_strength < OPS_TOLERANCE)
   return index_glob;
return index_diff;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::setOutputPixelAverage(int neighbour_count)
{
// average
int k;
float nbc = targetCentral.fStrength;
for (k = 0; k < neighbour_count; k++)
	{
	targetCentral.fBGR[0] += targetNbh[k].fBGR[0] * targetNbh[k].fStrength;
	targetCentral.fBGR[1] += targetNbh[k].fBGR[1] * targetNbh[k].fStrength;
	targetCentral.fBGR[2] += targetNbh[k].fBGR[2] * targetNbh[k].fStrength;
	nbc += targetNbh[k].fStrength;
	}
targetCentral.fBGR[0] /= nbc;
targetCentral.fBGR[1] /= nbc;
targetCentral.fBGR[2] /= nbc;
int index = getStrongestOne(neighbour_count);
targetCentral.fStrength = targetNbh[index].fStrength;
return index;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::setOutputPixelConvolve(int neighbour_count)
{
// convolve with Gauss strength kernel
targetCentral.fBGR[0] *= targetCentral.fStrength;
targetCentral.fBGR[1] *= targetCentral.fStrength;
targetCentral.fBGR[2] *= targetCentral.fStrength;
float nbc = targetCentral.fStrength, strength;
for (int k = 0; k < neighbour_count; k++)
	{
	strength = targetNbh[k].fStrength * N33_GAUSS[k];  //N33_STRENGTH[k];
	targetCentral.fBGR[0] += targetNbh[k].fBGR[0] * strength;
	targetCentral.fBGR[1] += targetNbh[k].fBGR[1] * strength;
	targetCentral.fBGR[2] += targetNbh[k].fBGR[2] * strength;
	nbc += strength;
	}
targetCentral.fBGR[0] /= nbc;
targetCentral.fBGR[1] /= nbc;
targetCentral.fBGR[2] /= nbc;
int index = getStrongestOne(neighbour_count);
targetCentral.fStrength = targetNbh[index].fStrength;
return index;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::setOutputPixelWeightedSum(int neighbour_count)
{
// weighted sum
//if (targetCentral.fStrength < OPS_TOLERANCE)
//   return setOutputPixelConvolve(neighbour_count);
// calculate quadrance from central pixel
targetCentral.fBGR[0] *= targetCentral.fStrength;
targetCentral.fBGR[1] *= targetCentral.fStrength;
targetCentral.fBGR[2] *= targetCentral.fStrength;
float nbc = targetCentral.fStrength;
for (int k = 0; k < neighbour_count; k++)
	{
	targetCentral.fBGR[0] += targetNbh[k].fBGR[0] * targetNbh[k].fStrength;
	targetCentral.fBGR[1] += targetNbh[k].fBGR[1] * targetNbh[k].fStrength;
	targetCentral.fBGR[2] += targetNbh[k].fBGR[2] * targetNbh[k].fStrength;
	nbc += targetNbh[k].fStrength;
	}
targetCentral.fBGR[0] /= nbc;
targetCentral.fBGR[1] /= nbc;
targetCentral.fBGR[2] /= nbc;
int index = getStrongestOne(neighbour_count);
targetCentral.fStrength = targetNbh[index].fStrength;
return index;
}
//---------------------------------------------------------------------------
int __fastcall TPaca::setOutputPixelStrongest(int neighbour_count)
{
// propagate strongest one, average strength
//if (neighbour_count < 8 || targetCentral.fStrength < OPS_TOLERANCE)
if (targetCentral.fStrength < OPS_TOLERANCE)
   return setOutputPixelConvolve(neighbour_count);
int index = getStrongestOne(neighbour_count);
if (targetCentral.fStrength > targetNbh[index].fStrength)
   return setOutputPixelConvolve(neighbour_count);
if (fabs(targetCentral.fStrength - targetNbh[index].fStrength) < OPS_TOLERANCE)
   return setOutputPixelConvolve(neighbour_count);
// two combo
/*
float nbc = targetCentral.fStrength +  targetNbh[index].fStrength;
targetCentral.fBGR[0] = (targetCentral.fBGR[0]*targetCentral.fStrength + targetNbh[index].fBGR[0] * targetNbh[index].fStrength) /nbc;
targetCentral.fBGR[1] = (targetCentral.fBGR[1]*targetCentral.fStrength + targetNbh[index].fBGR[1] * targetNbh[index].fStrength) /nbc;
targetCentral.fBGR[2] = (targetCentral.fBGR[2]*targetCentral.fStrength + targetNbh[index].fBGR[2] * targetNbh[index].fStrength) /nbc;
targetCentral.fStrength = nbc / 2.0;
*/
targetCentral.fBGR[0] = targetNbh[index].fBGR[0];
targetCentral.fBGR[1] = targetNbh[index].fBGR[1];
targetCentral.fBGR[2] = targetNbh[index].fBGR[1];
targetCentral.fStrength = targetNbh[index].fStrength;
return targetNbh[index].SiteIndex;
}
//---------------------------------------------------------------------------
void __fastcall TPaca::setOutputPixel(int i, int j, float dstWeight)
{
//
//if (maskSurface->DirectByte[i][j] < 0xff)
if (dstWeight < 1.0)
   {
   //float fByte = (float)maskSurface->DirectByte[i][j] / 255.0;
   dstSurface->DirectBGR[i][j].bgr[0] = (BYTE)fast_ftol(dstWeight * targetCentral.fBGR[0] + (1.0 - dstWeight)*(float)srcSurface->DirectBGR[i][j].bgr[0]);
   dstSurface->DirectBGR[i][j].bgr[1] = (BYTE)fast_ftol(dstWeight * targetCentral.fBGR[1] + (1.0 - dstWeight)*(float)srcSurface->DirectBGR[i][j].bgr[1]);
   dstSurface->DirectBGR[i][j].bgr[2] = (BYTE)fast_ftol(dstWeight * targetCentral.fBGR[2] + (1.0 - dstWeight)*(float)srcSurface->DirectBGR[i][j].bgr[2]);
   }
else
   {
   dstSurface->DirectBGR[i][j].bgr[0] = (BYTE)fast_ftol(targetCentral.fBGR[0]);
   dstSurface->DirectBGR[i][j].bgr[1] = (BYTE)fast_ftol(targetCentral.fBGR[1]);
   dstSurface->DirectBGR[i][j].bgr[2] = (BYTE)fast_ftol(targetCentral.fBGR[2]);
   }
}
//---------------------------------------------------------------------------
void __fastcall TPaca::processCA(void)  // parameter taret <-> source
{
// The core of complete story - cellular automaton with logic rules.
int i, j, neighbour_count, site_index;
//int si, sj;
bool coordsOk;
BYTE testByte = CTRL_MASKED | CTRL_PARTIALLY_MASKED;
//BYTE testByte = CTRL_MASKED;
targetSiteCreated = false;
sourceSiteCreated = false;
for (i = targetRect.top; i <= targetRect.bottom; i++)
	{
	for (j = targetRect.left; j <= targetRect.right; j++)
		{
		//if (controlInputSurface->DirectByte[i][j] != CTRL_MASKED)
		if (!(controlInputSurface->DirectByte[i][j] & testByte ))
		   continue;
		resetLogicContainer();
		neighbour_count = exploreTargetNeighbour(i, j);       // sets site strength and target coords
		if (neighbour_count == 0)
		   continue;
		if (doCreation(i, j, neighbour_count))
		   {
		   coordsOk = getSourceCoords(neighbour_count, CTRL_UNMASKED);
		   if (!coordsOk)
			  continue;
		   controlOutputSurface->DirectByte[i][j] = CTRL_TARGET;
		   targetSiteCreated = true;
		   targetSitesCount++;
		   site_index = getSiteIndex(i, j, neighbour_count);
		   // test
		   //dstSurface->DirectBGR[i][j].bgr[0] = 0x00;
		   //dstSurface->DirectBGR[i][j].bgr[1] = 0x00;
		   //dstSurface->DirectBGR[i][j].bgr[2] = 0x00;
		   controlOutputSurface->DirectByte[iCoordY][iCoordX] = CTRL_SOURCE;
		   targetCentral.fBGR[0] = (float)srcSurface->DirectBGR[iCoordY][iCoordX].bgr[0];
		   targetCentral.fBGR[1] = (float)srcSurface->DirectBGR[iCoordY][iCoordX].bgr[1];
		   targetCentral.fBGR[2] = (float)srcSurface->DirectBGR[iCoordY][iCoordX].bgr[2];
		   calculateStrength(i, j, neighbour_count, targetCentral.fBGR);
		   //
		   pacaMap.X[i][j] = iCoordX;
		   pacaMap.Y[i][j] = iCoordY;
		   pacaMap.SiteIndex[i][j] = site_index;
		   pacaMap.OutStrength[i][j] = targetCentral.fStrength;
		   if (targetCentral.fStrength > fMaxStrength)
			  fMaxStrength = targetCentral.fStrength;
		   if (maskSurface->DirectByte[i][j] < 0xff)
			  {
			  float fByte = (float)maskSurface->DirectByte[i][j] / 255.0;
			  dstSurface->DirectBGR[i][j].bgr[0] = (BYTE)fast_ftol(fByte * targetCentral.fBGR[0] + (1 - fByte)*(float)srcSurface->DirectBGR[i][j].bgr[0]);
			  dstSurface->DirectBGR[i][j].bgr[1] = (BYTE)fast_ftol(fByte * targetCentral.fBGR[1] + (1 - fByte)*(float)srcSurface->DirectBGR[i][j].bgr[1]);
			  dstSurface->DirectBGR[i][j].bgr[2] = (BYTE)fast_ftol(fByte * targetCentral.fBGR[2] + (1 - fByte)*(float)srcSurface->DirectBGR[i][j].bgr[2]);
			  }
		   else
			  {
			  dstSurface->DirectBGR[i][j].bgr[0] = srcSurface->DirectBGR[iCoordY][iCoordX].bgr[0];
			  dstSurface->DirectBGR[i][j].bgr[1] = srcSurface->DirectBGR[iCoordY][iCoordX].bgr[1];
			  dstSurface->DirectBGR[i][j].bgr[2] = srcSurface->DirectBGR[iCoordY][iCoordX].bgr[2];
		      }
		   // test
		   //dstSurface->DirectBGR[iCoordY][iCoordX].bgr[0] = 0x00;
		   //dstSurface->DirectBGR[iCoordY][iCoordX].bgr[1] = 0xff;
		   //dstSurface->DirectBGR[iCoordY][iCoordX].bgr[2] = 0x00;
		   sourceSiteCreated = true;
		   }
		}
	}
// copy out to in
//for (i = 0; i < imgHeight; i++)
for (i = sourceRect.top; i < sourceRect.bottom; i++)
	{
	//for (j = 0; j < imgWidth; j++)
	for (j = sourceRect.left; j < sourceRect.right; j++)
		{
		controlInputSurface->DirectByte[i][j] = controlOutputSurface->DirectByte[i][j];
		pacaMap.InpStrength[i][j] = pacaMap.OutStrength[i][j];
		//srcSurface->DirectBGR[i][j].bgr[0] = dstSurface->DirectBGR[i][j].bgr[0];
		//srcSurface->DirectBGR[i][j].bgr[1] = dstSurface->DirectBGR[i][j].bgr[1];
		//srcSurface->DirectBGR[i][j].bgr[2] = dstSurface->DirectBGR[i][j].bgr[2];
		}
	}
//controlOutput->CopyRectTo(controlInput, 0, 0, 0, 0, imgWidth, imgHeight);
maxProbability = fast_ftol(fMaxStrength);
}
//---------------------------------------------------------------------------
void __fastcall TPaca::copyDst2Src(void)
{
for (int i = targetRect.top; i < targetRect.bottom; i++)
	{
	for (int j = targetRect.left; j < targetRect.right; j++)
		{
		srcSurface->DirectBGR[i][j].bgr[0] = dstSurface->DirectBGR[i][j].bgr[0];
		srcSurface->DirectBGR[i][j].bgr[1] = dstSurface->DirectBGR[i][j].bgr[1];
		srcSurface->DirectBGR[i][j].bgr[2] = dstSurface->DirectBGR[i][j].bgr[2];
		}
	}
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::fillHoles(int **outSiteIndex)
{
// fill holes in inpainted area
int i, j, neighbour_count, site_index;
//int si, sj;
bool holeRemoved = true;
copyDst2Src();
while (holeRemoved)
	{
	holeRemoved = false;
	for (i = targetRect.top; i < targetRect.bottom; i++)
		{
		for (j = targetRect.left; j < targetRect.right; j++)
			{
			// find the state of target pixel
			if (controlInputSurface->DirectByte[i][j] != CTRL_MASKED)
			   continue;
			// explore neighbour around hole
			neighbour_count = exploreHoleNeighbour(i, j, CTRL_TARGET);      // sets site strength and target coords
			if (neighbour_count < 2)  // skip corners
			//if (!neighbour_count)
			   continue;
			site_index = doPixelOutput(neighbour_count);
			if (site_index == -1)
			   {
			   int mate = 23;
			   }
			setOutputPixel(i, j, 1.0);
			controlOutputSurface->DirectByte[i][j] =  CTRL_STITCHED;
			pacaMap.OutStrength[i][j] = targetCentral.fStrength;
			outSiteIndex[i][j] = site_index;
			if ( targetCentral.fStrength < OPS_TOLERANCE)
			   {
			   int mate = 23;
			   }
			holeRemoved = true;
			}
		}
	if (!holeRemoved)
	   return false;
	for (i = targetRect.top; i < targetRect.bottom; i++)
		{
		for (j = targetRect.left; j < targetRect.right; j++)
			{
			if (controlOutputSurface->DirectByte[i][j] == CTRL_STITCHED)
			   {
			   controlOutputSurface->DirectByte[i][j] = CTRL_TARGET;
			   controlInputSurface->DirectByte[i][j] = controlOutputSurface->DirectByte[i][j];
			   pacaMap.InpStrength[i][j] = pacaMap.OutStrength[i][j];
			   pacaMap.SiteIndex[i][j] = outSiteIndex[i][j];
			   srcSurface->DirectBGR[i][j].bgr[0] = dstSurface->DirectBGR[i][j].bgr[0];
			   srcSurface->DirectBGR[i][j].bgr[1] = dstSurface->DirectBGR[i][j].bgr[1];
			   srcSurface->DirectBGR[i][j].bgr[2] = dstSurface->DirectBGR[i][j].bgr[2];
			   }
		   }
	   }
	}
//controlOutput->CopyRectTo(controlInput, 0, 0, 0, 0, imgWidth, imgHeight);
return holeRemoved;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::combineSites(int **outSiteIndex, int iterations)
{
// Smooth coliding sites.
// We are interested only in sites with different indices.
int i, j, k, neighbour_count, site_index;
//int si, sj;
bool siteCombined = false;
copyDst2Src();
float smoothf = 1.0 / (iterations + 1);
for (int iter = 0; iter < iterations; iter++)
	{
	for (i = targetRect.top; i < targetRect.bottom; i++)
		{
		for (j = targetRect.left; j < targetRect.right; j++)
			{
			// find the state of target pixel
			if (controlInputSurface->DirectByte[i][j] != CTRL_TARGET)
			   continue;
			// explore neighbour around the site
			//neighbour_count = exploreSiteNeighbour(i, j, CTRL_TARGET | CTRL_STITCHED);      // sets site strength and target coords
			neighbour_count = exploreSiteNeighbour(i, j, CTRL_TARGET);      // sets site strength and target coords
			if (!neighbour_count)
			   continue;
			// skip neighbour where all sites have the same index
			k = neighbour_count;
			while(--k > 0 && targetNbh[k].SiteIndex == targetNbh[0].SiteIndex);
			if (k == 0)
			   continue;
			site_index = doPixelOutput(neighbour_count);
			if (site_index == -1)
			   continue;
			//if (site_index == pacaMap.SiteIndex[i][j])
			//   continue;
			setOutputPixel(i, j, 1.0 - (smoothf * (iter + 1.0)));
			controlOutputSurface->DirectByte[i][j] = CTRL_STITCHED;
			pacaMap.OutStrength[i][j] = targetCentral.fStrength;
			outSiteIndex[i][j] = site_index;
			// test
			//dstSurface->DirectBGR[i][j].bgr[0] = 0x00;
			//dstSurface->DirectBGR[i][j].bgr[1] = 0x00;
			//dstSurface->DirectBGR[i][j].bgr[2] = 0x00;
			siteCombined = true;
			}
		}
	if (!siteCombined)
	   return false;
	// copy out to in
	for (i = targetRect.top; i < targetRect.bottom; i++)
		{
		for (j = targetRect.left; j < targetRect.right; j++)
			{
			//pacaMap.InpStrength[i][j] = pacaMap.OutStrength[i][j];
			if (controlOutputSurface->DirectByte[i][j] == CTRL_STITCHED)
			   {
			   controlOutputSurface->DirectByte[i][j] = CTRL_TARGET;
			   controlInputSurface->DirectByte[i][j] = controlOutputSurface->DirectByte[i][j];
			   pacaMap.InpStrength[i][j] = pacaMap.OutStrength[i][j];
			   pacaMap.SiteIndex[i][j] = outSiteIndex[i][j];
			   srcSurface->DirectBGR[i][j].bgr[0] = dstSurface->DirectBGR[i][j].bgr[0];
			   srcSurface->DirectBGR[i][j].bgr[1] = dstSurface->DirectBGR[i][j].bgr[1];
			   srcSurface->DirectBGR[i][j].bgr[2] = dstSurface->DirectBGR[i][j].bgr[2];
			   }
		   }
	   }
	}
return siteCombined;
}
//---------------------------------------------------------------------------
bool __fastcall TPaca::smoothBorder(int sF)
{
// Smooth selection border. Use it only when selection feather is 0!!!
int i, j, k, neighbour_count, site_index, sfb;
//int si, sj;
bool siteCombined = false;
copyDst2Src();
TRect boundRect;
sfb = sF + 1;
boundRect.left  = targetRect.left - sfb ;
boundRect.right = targetRect.right + sfb;
boundRect.top   = targetRect.top - sfb;
boundRect.bottom   = targetRect.bottom + sfb;
if (boundRect.left < 0)
   boundRect.left = 0;
if (boundRect.right > imgWidth)
   boundRect.right = imgWidth;
if (boundRect.top < 0)
   boundRect.top = 0;
if (boundRect.bottom > imgHeight)
   boundRect.bottom = imgHeight;
//float smoothf = 255.0 / (sF + 1);
float smoothf = 1.0 / (sF + 1);
for (int iter = 0; iter < sF; iter++)
	{
	for (i = boundRect.top; i < boundRect.bottom; i++)
		{
		for (j = boundRect.left; j < boundRect.right; j++)
			{
			// find the state of target pixel
			if (controlInputSurface->DirectByte[i][j] & (CTRL_MASKED | CTRL_PARTIALLY_MASKED | CTRL_TARGET))
			   continue;
			// explore neighbour around border
			neighbour_count = exploreBorderNeighbour(i, j, CTRL_TARGET);      // sets site strength and target coords
			//neighbour_count = exploreSiteNeighbour(i, j, CTRL_TARGET);      // sets site strength and target coords
			if (!neighbour_count)
			   continue;
			//setOutputPixelAverage(neighbour_count);	// gut ja!
			setOutputPixelConvolve(neighbour_count);	// gut ja!
			setOutputPixel(i, j, 1.0 - (smoothf * (iter + 1.0)));
			controlOutputSurface->DirectByte[i][j] = CTRL_STITCHED;
			}
		}
	// copy out to in
	for (i = boundRect.top; i < boundRect.bottom; i++)
		{
		for (j = boundRect.left; j < boundRect.right; j++)
			{
			//pacaMap.InpStrength[i][j] = pacaMap.OutStrength[i][j];
			if (controlOutputSurface->DirectByte[i][j] == CTRL_STITCHED)
			   {
			   controlOutputSurface->DirectByte[i][j] = CTRL_TARGET;
			   controlInputSurface->DirectByte[i][j] = controlOutputSurface->DirectByte[i][j];
			   srcSurface->DirectBGR[i][j].bgr[0] = dstSurface->DirectBGR[i][j].bgr[0];
			   srcSurface->DirectBGR[i][j].bgr[1] = dstSurface->DirectBGR[i][j].bgr[1];
			   srcSurface->DirectBGR[i][j].bgr[2] = dstSurface->DirectBGR[i][j].bgr[2];
			   }
		   }
	   }
	}
return true;
}
//---------------------------------------------------------------------------
// Post processing
//---------------------------------------------------------------------------
bool __fastcall TPaca::PostProcess(bool pFillHoles, int pPropagateStrength, int pSelectionFeather)
{
// Eestore original rectangle (just in case we have tempered somthing - old onion peeling version).
targetRect.left  = selRect->left;
targetRect.right = selRect->right;
targetRect.top   = selRect->top;
targetRect.bottom = selRect->bottom;
int **outSiteIndex = new int *[imgHeight];
for (int i = 0; i < imgHeight; i++)
   {
   outSiteIndex[i] = new int [imgWidth];
   for (int j = 0; j < imgWidth; j++)
	   outSiteIndex[i][j] = -1;
   }
// fill unoccupied areas (holes)
if (pFillHoles)
   fillHoles(outSiteIndex);
if (pPropagateStrength)
   combineSites(outSiteIndex, pPropagateStrength);
if (pSelectionFeather)
   smoothBorder(pSelectionFeather);
for (int i = 0; i < imgHeight; i++)
   delete []outSiteIndex[i];
delete []outSiteIndex;
return true;
}
//---------------------------------------------------------------------------






