//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "seed.h"
#include <math.h>
#include <algorithm>
#include "ASMVars.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
// Pantoghraphic Cellular Automata Inpainting
//
// Seeding class
//
// Author: Sinia Petri
// You may freely distribute, copy, cut and paste this code, as long as...whatever.
//---------------------------------------------------------------------------
__fastcall TSeed::TSeed(TSpadeSurfaceBasic *pMaskSurface, TSpadeSurfaceBasic *pControlSurface, TRect *pMaskRect, BYTE pHardMaskByte)
{
maskSurface = pMaskSurface;
controlSurface = pControlSurface;
targetRect = pMaskRect;
hardMaskByte = pHardMaskByte;
outputPts  = new TObjectList(false);
}
//---------------------------------------------------------------------------
// destructor
__fastcall TSeed::~TSeed(void)
{
Clear();
delete outputPts;
outputPts = 0;
}
//---------------------------------------------------------------------------
void __fastcall TSeed::Clear(void)
{
while (outputPts->Count > 0)
	{
	TSeedPoint *outPt = (TSeedPoint*)(outputPts->Items[outputPts->Count-1]);
	delete outPt;
	outputPts->Delete(outputPts->Count-1);
	}
}
//---------------------------------------------------------------------------
bool __fastcall TSeed::Seed(int pSeedStyle, int pDivision)
{
cellX = targetRect->Width()  / (float)pDivision;
cellY = targetRect->Height() / (float)pDivision;
if (cellX < 1 || cellY < 1)
   return false;
gridDivision = pDivision;
switch (pSeedStyle)
  {
  case SEED_EXTREMA_H:
	   SweepSelection(sHorizontal);
	   break;
  case SEED_EXTREMA_V:
	   SweepSelection(sVertical);
	   break;
  case SEED_EXTREMA_HV:
	   SweepSelection(sBoth);
	   break;
  case SEED_GRID_NOC:
  case SEED_GRID_FULL:
	   gridSeed (false);	//TODO: exclude corners
	   break;
  default:
	   break;
  }
// set walk direction (oposite to sweep direction)
for (int k = 0; k < outputPts->Count; k++)
	{
	TSeedPoint *outPt = (TSeedPoint*)(outputPts->Items[k]);
	switch (outPt->sweepDirection)
	   {
	   case wRight:
			outPt->walkDirection = wLeft;
			break;
	   case wLeft:
			outPt->walkDirection = wRight;
			break;
	   case wDown:
			outPt->walkDirection = wUp;
			break;
	   case wUp:
			outPt->walkDirection = wDown;
			break;
	   }
	}
return (bool)(outputPts->Count);
}
//---------------------------------------------------------------------------
void __fastcall TSeed::gridSeed(bool excludeCorners)
{
TPoint seed;
float fi, fj;
int i, j, k, seedX, seedY;
int nCellY = fast_ftol(cellY);
int nCellX = fast_ftol(cellX);
TRect boundRect;
TSeedPoint *outPt;
// eventual border adjustment
boundRect.left    = targetRect->left;
boundRect.right   = targetRect->right;
boundRect.top     = targetRect->top;
boundRect.bottom  = targetRect->bottom;
// let's find left border pts
for (fi = boundRect.top; fi < boundRect.bottom; fi += cellY)
	{
	i = fast_ftol(fi);
	seedY = i + (nCellY>>1);	// middle of the row
	for (j = boundRect.left; j < boundRect.right; j ++)
		{
		seedX = j;
		if (!controlSurface->SetSafeIndex(&seedY, &seedX))
		   continue;	// bad point
		if (controlSurface->DirectByte[seedY][seedX] & SD_MASKED)	// got it
		   {
		   outPt = new TSeedPoint;
		   outPt->x = seedX;
		   outPt->y = seedY;
		   outPt->sweepDirection = wRight;
		   outputPts->Add((TObject *)outPt);		   
		   break;
		   }
		}
	}
// let's find right border pts
for (fi = boundRect.top; fi < boundRect.bottom; fi += cellY)
	{
	i = fast_ftol(fi);
	seedY = i + (nCellY>>1);	// middle of the row
	for (j = boundRect.right-1; j >= boundRect.left; j --)
		{
		seedX = j;
		if (!controlSurface->SetSafeIndex(&seedY, &seedX))
		   continue;	// bad point
		if (controlSurface->DirectByte[seedY][seedX] & SD_MASKED)	// got it
		   {
		   outPt = new TSeedPoint;
		   outPt->x = seedX;
		   outPt->y = seedY;
		   outPt->sweepDirection = wLeft;
		   outputPts->Add((TObject *)outPt);		   
		   break;
		   }
		}
	}
if (gridDivision == 2)
   return;
// let's find up border pts
for (fj = boundRect.left + cellX; fj < (boundRect.right - cellX); fj += cellX)
	{
	j = fast_ftol(fj);
	seedX = j + (nCellX>>1);	// middle of the column
	for (i = boundRect.top; i < boundRect.bottom; i++)
		{
		seedY = i;
		if (!controlSurface->SetSafeIndex(&seedY, &seedX))
		   continue;	// bad point
		if (controlSurface->DirectByte[seedY][seedX] & SD_MASKED)	// got it
		   {
		   outPt = new TSeedPoint;
		   outPt->x = seedX;
		   outPt->y = seedY;
		   outPt->sweepDirection = wDown;
		   outputPts->Add((TObject *)outPt);		   
		   break;
		   }
		}
	}
// let's find down border pts
for (fj = boundRect.left + cellX; fj < (boundRect.right - cellX); fj += cellX)
	{
	j = fast_ftol(fj);
	seedX = j + (nCellX>>1);	// middle of the column
	for (i = boundRect.bottom-1; i >= boundRect.top; i--)
		{
		seedY = i;
		if (!controlSurface->SetSafeIndex(&seedY, &seedX))
		   continue;	// bad point
		if (controlSurface->DirectByte[seedY][seedX] & SD_MASKED)	// got it
		   {
		   outPt = new TSeedPoint;
		   outPt->x = seedX;
		   outPt->y = seedY;
		   outPt->sweepDirection = wUp;
		   outputPts->Add((TObject *)outPt);		   
		   break;
		   }
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TSeed::SweepSelection(TSweepType type)
{
switch (type)
	{
	case sVertical:
		sweepMask(wDown);
		sweepMask(wUp);
		break;
	case sHorizontal:
		sweepMask(wRight);
		sweepMask(wLeft);
		break;
	default:
		sweepMask(wDown);
		sweepMask(wUp);
		sweepMask(wRight);
		sweepMask(wLeft);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TSeed::sweepMask(TWalkDirection dir)
{
// sweep mask - assume no holes in it
int i, j, k, sl;
bool *sweepLine, sweepPointFound, segOut;
int segStart, segEnd, lp, lpIncr, swl, sweepLineFull;
TObjectList *extremaPts = new TObjectList(false);
if (dir == wRight || dir == wLeft)
   {
   sweepLine = new bool[targetRect->Height()];
   sweepLineCondition = sweepDown;
   swl = targetRect->top;
   if (dir == wRight)
	  {
	  lp = targetRect->left;
	  lpIncr = +1;
	  sweepLoopCondition = sweepRight;
	  }
   else
	  {
	  lp = targetRect->right-1;
	  lpIncr = -1;
	  sweepLoopCondition = sweepLeft;
	  }
   for (i = targetRect->top, sl ; i < targetRect->bottom; i++)
	   sweepLine[i - swl] = false;
   // left and right extrema
   j = lp;
   while (sweepLoopCondition(j))
	   {
	   sweepPointFound = false;
	   sweepLineFull = true;
	   for (i = targetRect->top; i < targetRect->bottom; i++)
		   {
		   if (maskSurface->DirectByte[i][j] >= hardMaskByte)
			  {
			  sweepLine[i - swl] = true;
			  sweepPointFound = true;
			  }
		   sweepLineFull &= sweepLine[i - swl];
		   }
	   if (!sweepPointFound)
		  {
		  j += lpIncr;
		  continue;
		  }
	   // let's examine the swep line
	   sweepLineExam(dir, swl, j, sweepLine, extremaPts);
	   j += lpIncr;
	   if (sweepLineFull)
		  break;
	   }
   }
else
   {
   sweepLine = new bool[targetRect->Width()];
   sweepLineCondition = sweepRight;
   swl = targetRect->left;
   if (dir == wDown)
	  {
	  lp = targetRect->top;
	  lpIncr = +1;
	  sweepLoopCondition = sweepDown;
	  }
   else
	  {
	  lp = targetRect->bottom-1;
	  lpIncr = -1;
	  sweepLoopCondition = sweepUp;
	  }
   for (j = targetRect->left; j < targetRect->right; j++)
	   sweepLine[j - swl] = false;
   // down up extrema
   i = lp;
   while (sweepLoopCondition(i))
	   {
	   sweepPointFound = false;
	   sweepLineFull = true;
	   for (j = targetRect->left; j < targetRect->right; j++)
		   {
		   if (maskSurface->DirectByte[i][j] >= hardMaskByte)
			  {
			  sweepLine[j - swl] = true;
			  sweepPointFound = true;
			  }
		   sweepLineFull &= sweepLine[j - swl];
		   }
	   if (!sweepPointFound)
		  {
		  i += lpIncr;
		  continue;
		  }
	   // let's examine the swep line
	   sweepLineExam(dir, swl, i, sweepLine, extremaPts);
	   i += lpIncr;
	   if (sweepLineFull)
		  break;
	   }
   }
delete []sweepLine;
for (k = 0; k < extremaPts->Count; k++)
	{
	TSeedPoint *inPt = (TSeedPoint*)(extremaPts->Items[k]);
	TSeedPoint *outPt = new TSeedPoint;
	outPt->x = inPt->x;
	outPt->y = inPt->y;
	outPt->sweepDirection = inPt->sweepDirection;
	outPt->walkDirection  = inPt->walkDirection;
	outputPts->Add((TObject*)outPt);
	}
while (extremaPts->Count > 0)
	{
	TSeedPoint *inPt = (TSeedPoint*)(extremaPts->Items[extremaPts->Count-1]);
	delete inPt;
	extremaPts->Delete(extremaPts->Count-1);
	}
delete extremaPts;
extremaPts = 0;
}
//---------------------------------------------------------------------------
void __fastcall TSeed::sweepLineExam(TWalkDirection dir, int lp, int fix, bool *sweepLine, TObjectList *extremaPts)
{
// let's examine the swep line
int segStart = -1, segEnd = -1;
bool segOut = true, ptInside;
int swl = lp;
TSeedPoint *outPt;
while (sweepLineCondition(lp))
   {
   if (sweepLine[lp - swl])
	  {
	  if (segStart == -1)
		 {
		 segStart = lp;
		 segEnd   = lp;
		 segOut = false;
		 }
	  else
		 segEnd = lp;
	  }
   else // check segment
	  {
	  if (segStart != -1)
		 {
		 // check if one of the points is inside the segment
		 for (int k = 0; k < extremaPts->Count; k++)
			 {
			 outPt = (TSeedPoint*)(extremaPts->Items[k]);
			 if (dir == wRight || dir == wLeft)
				ptInside = (outPt->y >= segStart && outPt->y <= segEnd);
			 else
				ptInside = (outPt->x >= segStart && outPt->x <= segEnd);
			 if (ptInside)
				{
				segStart = -1;
				segOut = true;
				break;
				}
			 }
		 if (segStart == -1)
			{
			lp++;
			continue;
			}
		 // new segment
		 // calculate middle point
		 outPt = new TSeedPoint;
		 outPt->sweepDirection = dir;
		 if (dir == wRight || dir == wLeft)
			{
			outPt->x = fix;
			outPt->y = (segEnd + segStart) / 2;
			}
		 else
			{
			outPt->x = (segEnd + segStart) / 2;
			outPt->y = fix;
			}
		 // pump out the point
		 extremaPts->Add((TObject *)outPt);
		 segStart = -1;
		 segOut = true;
		 }
	  }
   lp++;
   }
   // check if last segment is out (selection may be a rectangle)
   if (!segOut)
	  {
	  // check if one of the points is inside the segment
	  for (int k = 0; k < extremaPts->Count; k++)
		  {
		  outPt = (TSeedPoint*)(extremaPts->Items[k]);
		  if (dir == wRight || dir == wLeft)
			 ptInside = (outPt->y >= segStart && outPt->y <= segEnd);
		  else
			 ptInside = (outPt->x >= segStart && outPt->x <= segEnd);
		  if (ptInside)
			 return;
		 }
	  // calculate middle point
	  outPt = new TSeedPoint;
	  outPt->sweepDirection = dir;
	  if (dir == wRight || dir == wLeft)
		 {
		 outPt->x = fix;
		 outPt->y = (segEnd + segStart) / 2;
		 }
	  else
		 {
		 outPt->x = (segEnd + segStart) / 2;
		 outPt->y = fix;
		 }
	  // pump out the point
	  extremaPts->Add((TObject *)outPt);
	  segStart = -1;
	  }
}
//---------------------------------------------------------------------------
bool __fastcall TSeed::sweepRight(int i)
{
return i < targetRect->right;
}
//---------------------------------------------------------------------------
bool __fastcall TSeed::sweepLeft(int i)
{
return i >= targetRect->left;
}
//---------------------------------------------------------------------------
bool __fastcall TSeed::sweepDown(int i)
{
return i < targetRect->bottom;
}
//---------------------------------------------------------------------------
bool __fastcall TSeed::sweepUp(int i)
{
return i >= targetRect->top;
}


